/**************************************************************************
*  This file is part of the TAL project (Tiny Abstraction Layer)
*
*  Copyright (c) 2016 by Michael Fischer (www.emb4fun.de).
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without 
*  modification, are permitted provided that the following conditions 
*  are met:
*  
*  1. Redistributions of source code must retain the above copyright 
*     notice, this list of conditions and the following disclaimer.
*  2. Redistributions in binary form must reproduce the above copyright
*     notice, this list of conditions and the following disclaimer in the 
*     documentation and/or other materials provided with the distribution.
*  3. Neither the name of the author nor the names of its contributors may 
*     be used to endorse or promote products derived from this software 
*     without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
*  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
*  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
*  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
*  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
*  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
*  SUCH DAMAGE.
*
***************************************************************************
*  History:
*
*  23.02.2016  mifi  First Version, tested with an Embedded Artists 
*                    LPC1788 Developer's Kit.
**************************************************************************/
#define __TALCPU_COM_C__

/*=======================================================================*/
/*  Include                                                              */
/*=======================================================================*/
#include "tal.h"

/*=======================================================================*/
/*  All Structures and Common Constants                                  */
/*=======================================================================*/

/*
 * Use UART FIFO
 *
 *    0 = disabled
 *    1 = enabled
 */
#define USE_UART_FIFO   1

/*=======================================================================*/
/*  Definition of all local Data                                         */
/*=======================================================================*/

static TAL_COM_DCB *DCBArray[TAL_COM_PORT_MAX];

/*=======================================================================*/
/*  Definition of all local Procedures                                   */
/*=======================================================================*/

/*************************************************************************/
/*  ReadData                                                             */
/*                                                                       */
/*  Read data from the Receive Buffer Register.                          */
/*                                                                       */
/*  In    : pDCB, pUARTx                                                 */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void ReadData (TAL_COM_DCB *pDCB, LPC_USART_T *pUARTx)
{
   TAL_RESULT  Error;
   uint8_t    bData;

   /* Read data */
   bData = Chip_UART_ReadByte(pUARTx);
   
   /* If we have no overflow... */
   if (TAL_FALSE == pDCB->bRxOverflow)
   {
      /* ... put it into the ring buffer */
      Error = tal_MISCRingAdd(&pDCB->RxRing, &bData);
      if (TAL_OK == Error)
      {
         /* Signal counting semaphore */
         OSSemaSignalFromInt(&pDCB->RxRdySema);
      }
      else
      {
         /* Ups, overflow */
         pDCB->bRxOverflow = TAL_OK;
      }
   }  

} /* ReadData */

/*************************************************************************/
/*  IRQHandler                                                           */
/*                                                                       */
/*  This is the generic IRQ handler.                                     */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void IRQHandler (TAL_COM_DCB *pDCB)
{
   TAL_RESULT      Error;
   LPC_USART_T   *pUARTx;   
   uint32_t       dIIR;
   uint32_t       dIER;
   uint32_t       dLSR;
   uint8_t        bIndex;
   uint8_t        bData;

   /* Get USART pointer */
   pUARTx = (LPC_USART_T *)pDCB->HW.dBaseAddress;

   /* Get IRQ status */   
   dIIR = Chip_UART_ReadIntIDReg(pUARTx);
   dLSR = Chip_UART_ReadLineStatus(pUARTx);
   dIER = Chip_UART_GetIntsEnabled(pUARTx);
   
   
   /* 
    * Check for RX interrupt
    */
    
   /* Check if FIFO is used */ 
   if (pDCB->HW.bFifoTriggerLevel != 0)
   {
      /*
       * FIFO is used
       */
       
      /* Check for a block of data, FIFO trigger level */
      if      (UART_IIR_INTID_RDA == (dIIR & UART_IIR_INTID_MASK))
      {
         /* A block of data, FIFO trigger level,  can be read */
         for (bIndex = 0; bIndex < pDCB->HW.bFifoTriggerLevel; bIndex++)
         {
            ReadData(pDCB, pUARTx);
         }
      }
      /* Check Character Time-out Indicator */
      else if (UART_IIR_INTID_CTI == (dIIR & UART_IIR_INTID_MASK))
      {
         while (dLSR & UART_LSR_RDR)
         {
            ReadData(pDCB, pUARTx);
            dLSR = Chip_UART_ReadLineStatus(pUARTx);
         }
      }
   }
   else
   {
      /*
       * FIFO is NOT used
       */
       
      /* Check if data is available */
      if (UART_IIR_INTID_RDA == (dIIR & UART_IIR_INTID_MASK))
      {
         ReadData(pDCB, pUARTx);
      } /* end "RX interrupt" */
   }      
   
   
   /* 
    * Check for TX interrupt, but only if enabled 
    */
   if ((dIER & UART_IER_THREINT) &&                            /* <= enabled ? */
       (UART_IIR_INTID_THRE == (dIIR & UART_IIR_INTID_MASK)))  /* <= TX interrupt ? */
   {
      /* Read Data from the ring buffer */
      Error = tal_MISCRingGet(&pDCB->TxRing, &bData);
      if (Error != TAL_OK)
      {
         /* Ups, no data available, disable interrupt */
         Chip_UART_IntDisable(pUARTx, UART_IER_THREINT);
      }
      else
      {
         /* Check if FIFO is used */ 
         if (pDCB->HW.bFifoTriggerLevel != 0)
         {
            /*
             * FIFO is used
             */
             
            /* Put first data into the FIFO */
            Chip_UART_SendByte(pUARTx, bData);
            
            /* Get next data, sub one because the first data is in the FIFO */   
            bIndex = pDCB->HW.bFifoTriggerLevel - 1;
            while (bIndex != 0)
            {
               /* Get next data */
               Error = tal_MISCRingGet(&pDCB->TxRing, &bData);
               if (Error != TAL_OK)
               {
                  /* No data available anymore in the TX ring buffer */
                  break;
               }
               else
               {
                  /* Put the data into the FIFO */
                  Chip_UART_SendByte(pUARTx, bData);
               }
               bIndex--;
            }
         }
         else
         {
            /*
             * FIFO is NOT used
             */
            Chip_UART_SendByte(pUARTx, bData);
         }            
      }   
   } /* end "TX interrupt */
   
} /* IRQHandler */

/*=======================================================================*/
/*  All code exported                                                    */
/*=======================================================================*/

/*************************************************************************/
/*  USARTx_IRQHandler                                                    */
/*                                                                       */
/*  This is the Cortex USARTx IRQ handler.                               */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void UART0_IRQHandler (void)
{
   TAL_CPU_IRQ_ENTER();
   IRQHandler(DCBArray[TAL_COM_PORT_1]);
   TAL_CPU_IRQ_EXIT();
} /* UART0_IRQHandler */

void UART1_IRQHandler (void)
{
   TAL_CPU_IRQ_ENTER();
   IRQHandler(DCBArray[TAL_COM_PORT_2]);
   TAL_CPU_IRQ_EXIT();
} /* UART1_IRQHandler */

void UART2_IRQHandler (void)
{
   TAL_CPU_IRQ_ENTER();
   IRQHandler(DCBArray[TAL_COM_PORT_3]);
   TAL_CPU_IRQ_EXIT();
} /* UART2_IRQHandler */

void UART3_IRQHandler (void)
{
   TAL_CPU_IRQ_ENTER();
   IRQHandler(DCBArray[TAL_COM_PORT_4]);
   TAL_CPU_IRQ_EXIT();
} /* UART3_IRQHandler */

void UART4_IRQHandler (void)
{
   TAL_CPU_IRQ_ENTER();
   IRQHandler(DCBArray[TAL_COM_PORT_5]);
   TAL_CPU_IRQ_EXIT();
} /* UART4_IRQHandler */

/*************************************************************************/
/*  cpu_COMInit                                                          */
/*                                                                       */
/*  Prepare the hardware for use by the Open function later. Set the HW  */
/*  information depending of ePort and "enable" the COM port.            */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: TAL_OK / error cause                                         */
/*************************************************************************/
TAL_RESULT cpu_COMInit (TAL_COM_DCB *pDCB)
{
   TAL_RESULT    Error = TAL_ERR_COM_PORT_RANGE;
   TAL_COM_HW  *pHW    = &pDCB->HW;
   
   switch (pDCB->ePort)
   {
      case TAL_COM_PORT_1:
      {
         Error = tal_BoardEnableCOM1();
         if (TAL_OK == Error)
         {
            DCBArray[TAL_COM_PORT_1] = pDCB;
         
            pHW->dBaseAddress = LPC_UART0_BASE;
            pHW->IrqNumber    = UART0_IRQn;
            pHW->bIrqPriority = UART0_PRIO;

            /* Reset UART */
            Chip_SYSCTL_PeriphReset(SYSCTL_RESET_UART0);
         }
         break;
      } /* TAL_COM_PORT_1 */
     
      case TAL_COM_PORT_3:
      {
         Error = tal_BoardEnableCOM3();
         if (TAL_OK == Error)
         {
            DCBArray[TAL_COM_PORT_3] = pDCB;
         
            pHW->dBaseAddress = LPC_UART2_BASE;
            pHW->IrqNumber    = UART2_IRQn;
            pHW->bIrqPriority = UART2_PRIO;

            /* Reset UART */
            Chip_SYSCTL_PeriphReset(SYSCTL_RESET_UART2);
         }
         break;
      } /* TAL_COM_PORT_3 */
      
      default:
      {
         /* Do nothing */
         break;
      }
   } /* end switch (pDCB->ePort) */
   
   return(Error);
} /* cpu_COMInit */

/*************************************************************************/
/*  cpu_COMOpen                                                          */
/*                                                                       */
/*  Open the COM port.                                                   */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: TAL_OK / error cause                                         */
/*************************************************************************/
TAL_RESULT cpu_COMOpen (TAL_COM_DCB *pDCB)
{
   TAL_RESULT      Error = TAL_ERROR;
   LPC_USART_T   *pUARTx;   
   uint32_t       dLCR   = 0;
   uint8_t        bDummy = 0;
   
   /* Disable FIFO first */   
   pDCB->HW.bFifoTriggerLevel = 0;
   
   /* Get USART pointer */
   pUARTx = (LPC_USART_T *)pDCB->HW.dBaseAddress;
         
   /* Check word length */
   switch (pDCB->Settings.eLength)
   {
      case TAL_COM_LENGTH_8:
      {
         dLCR |= UART_LCR_WLEN8;
         break;
      }
      
      default:
      {
         Error = TAL_ERR_COM_LENGTH;
         goto COMOpenEnd;  /*lint !e801*/
         break;   /*lint !e527*/
      }
   } /* switch (pDCB->Settings.eLength) */
   
   /* Check parity settings */
   switch (pDCB->Settings.eParity)
   {
      case TAL_COM_PARITY_NONE:
      {
         dLCR |= UART_LCR_PARITY_DIS;  /*lint !e845 */
         break;
      } 
      
      case TAL_COM_PARITY_EVEN:
      {
         dLCR |= UART_LCR_PARITY_EVEN;
         break;
      }
      
      case TAL_COM_PARITY_ODD:
      {
         dLCR |= UART_LCR_PARITY_ODD;  /*lint !e845 */
         break;
      }
      
      default:
      {
         Error = TAL_ERR_COM_PARITY;
         goto COMOpenEnd;  /*lint !e801*/
         break;   /*lint !e527*/
      }
   } /* switch (pDCB->Settings.eParity) */
    
   /* Check stop bit settings */
   switch (pDCB->Settings.eStop)
   {
      case TAL_COM_STOP_1_0:
      {
         dLCR |= UART_LCR_SBS_1BIT;    /*lint !e845 */
         break;
      }
      
      case TAL_COM_STOP_2_0:
      {
         dLCR |= UART_LCR_SBS_2BIT;
         break;
      }
      
      default:
      {
         Error = TAL_ERR_COM_STOP;
         goto COMOpenEnd;  /*lint !e801*/
         break;   /*lint !e527*/
      }
   } /* switch (pDCB->Settings.eStop) */

   /* Check baud rate */
   if (pDCB->Settings.dBaudrate != 0)
   {
      Chip_UART_SetBaudFDR(pUARTx, pDCB->Settings.dBaudrate);
   }
   else
   {
      Error = TAL_ERR_COM_BAUDRATE;
      goto COMOpenEnd;  /*lint !e801*/
   }

   /* Enable transmitter and receiver */
   Chip_UART_TXEnable(pUARTx);
   
   /* Configure USART */
   Chip_UART_ConfigData(pUARTx, dLCR);
   
#if (USE_UART_FIFO >= 1)
   /* Reset and enable FIFOs, FIFO trigger level 2 (8 chars) */
   pDCB->HW.bFifoTriggerLevel = 8;                        
   Chip_UART_SetupFIFOS(pUARTx, (UART_FCR_FIFO_EN | 
                                 UART_FCR_RX_RS   | 
                                 UART_FCR_TX_RS   | 
                                 UART_FCR_TRG_LEV2) );
#endif                        
                        
   /* Dummy read to clear the RBR register */
   while (UART_LSR_RDR & Chip_UART_ReadLineStatus(pUARTx))
   {
      bDummy = Chip_UART_ReadByte(pUARTx);
   }  
   (void)bDummy;
   
   /* Enable RX interrrupt */
   Chip_UART_IntEnable(pUARTx, UART_IER_RBRINT);
   
   /* Set priority and enable the IRQ */
   tal_CPUIrqSetPriority(pDCB->HW.IrqNumber, pDCB->HW.bIrqPriority);
   tal_CPUIrqEnable(pDCB->HW.IrqNumber);

   Error = TAL_OK;
   
COMOpenEnd:   
   return(Error);
} /* cpu_COMOpen */

/*************************************************************************/
/*  cpu_COMClose                                                         */
/*                                                                       */
/*  Close the COM port.                                                  */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: TAL_OK / error cause                                         */
/*************************************************************************/
TAL_RESULT cpu_COMClose (TAL_COM_DCB *pDCB)
{
   TAL_RESULT    Error = TAL_OK;
   LPC_USART_T *pUARTx;   
   
   /* Get UART pointer */
   pUARTx = (LPC_USART_T *)pDCB->HW.dBaseAddress;

   /* Disable the IRQ */   
   tal_CPUIrqDisable(pDCB->HW.IrqNumber);
   
   /* Disable TX and RX interrupt  */
   Chip_UART_IntDisable(pUARTx, (UART_IER_RBRINT | UART_IER_THREINT));

   /* Disable the UART itself */
   Chip_UART_DeInit(pUARTx);
   
   return(Error);
} /* cpu_COMClose */

/*************************************************************************/
/*  cpu_COMStartTx                                                       */
/*                                                                       */
/*  Send the data from the ring buffer if the TX interrupt is disabled.  */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: TAL_OK / error cause                                         */
/*************************************************************************/
TAL_RESULT cpu_COMStartTx (TAL_COM_DCB *pDCB)
{
   TAL_RESULT    Error = TAL_OK;
   LPC_USART_T *pUARTx;   
   uint8_t      bData;
   
   /* Get USART pointer */
   pUARTx = (LPC_USART_T *)pDCB->HW.dBaseAddress;

   TAL_CPU_DISABLE_ALL_INTS();
   if (Chip_UART_GetIntsEnabled(pUARTx) & UART_IER_THREINT)
   {
      /* TX interrupt is enabled, do nothing */
   }
   else
   {
      /* Get data from the ring buffer */
      Error = tal_MISCRingGet(&pDCB->TxRing, &bData);
      if (TAL_OK == Error)
      {
         /* Send data */
         Chip_UART_SendByte(pUARTx, bData);
      
         /* Enable TX interrupt */
         Chip_UART_IntEnable(pUARTx, UART_IER_THREINT);
      }               
   }
   TAL_CPU_ENABLE_ALL_INTS();
   
   return(Error);
} /* cpu_COMStartTx */

/*************************************************************************/
/*  cpu_COMTxIsRunning                                                   */
/*                                                                       */
/*  Check if TX is still running.                                        */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: TAL_OK / TAL_ERROR                                           */
/*************************************************************************/
TAL_RESULT cpu_COMTxIsRunning (TAL_COM_DCB *pDCB)
{
   TAL_RESULT    Error = TAL_OK;
   LPC_USART_T *pUARTx;   
   
   /* Get USART pointer */
   pUARTx = (LPC_USART_T *)pDCB->HW.dBaseAddress;

   TAL_CPU_DISABLE_ALL_INTS();
   if (Chip_UART_GetIntsEnabled(pUARTx) & UART_IER_THREINT)
   {
      /* TX is still running */
      Error = TAL_OK;
   }
   else
   {
      /* TX is not running */
      Error = TAL_OK;
   }
   TAL_CPU_ENABLE_ALL_INTS();
   
   return(Error);
} /* cpu_COMTxIsRunning */

/*** EOF ***/
